# rtl_433
# Copyright (C) 2021 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime, timezone
import dateutil.parser
import dateutil.tz
import dateutil.utils
import json
import logging

METRICS = {
    "prom433_last_message": {}
}

HELP_FORMAT = "#HELP %s %s"
TYPE_FORMAT = "#TYPE %s %s "
METRIC_FORMAT = "%s{%s} %f"

TEMP_HELP = \
    "The temperature in degrees celcius."
TEMP_TYPE = "gauge"

HUMIDITY_HELP = "The humidity in %."
HUMIDITY_TYPE = "gauge"

WIND_AVG_HELP = "The average windspeed in km/h."
WIND_AVG_HELP_M = "The average windspeed in m/s."
WIND_AVG_TYPE = "gauge"
WIND_MAX_HELP = "The maximum windspeed in km/h."
WIND_MAX_HELP_M = "The maximum windspeed in m/s."
WIND_MAX_TYPE = "gauge"
WIND_DIR_HELP = "The wind direction in degrees."
WIND_DIR_TYPE = "gauge"

UV_HELP = "UV intensity raw value."
UV_TYPE = "gauge"
UVI_HELP = "UV Index."
UVI_TYPE = "gauge"
LUX_HELP = "Light LUX in W/m^2."
LUX_TYPE = "gauge"

RAIN_HELP = "The total rainfall in mm."
RAIN_TYPE = "counter"

BATTERY_HELP = "The battery status."
BATTERY_TYPE = "gauge"

BATTERY_V_HELP = "The battery voltage."
BATTERY_V_TYPE = "gauge"

SUPERCAP_V_HELP = "The supercapacitor voltage."
SUPERCAP_V_TYPE = "gauge"

LAST_MESSAGE_HELP = "The time the last message was received."
LAST_MESSAGE_TYPE = "counter"

FREQ_HELP = "The frequency the last message was received on."
FREQ_TYPE = "guage"

RSSI_HELP = "The RSSI of the last message."
RSSI_TYPE = "guage"

SNR_HELP = "The Signal to noise ratio of the last message."
SNR_TYPE = "guage"

NOISE_HELP = "The noise level of the last message."
NOISE_TYPE = "guage"

RADIO_CLOCK_HELP = "The radio clock value of the last message, in unix time."
RADIO_CLOCK_TYPE = "counter"

FIRMWARE_HELP = "The firmware version."
FIRMWARE_TYPE = "gauge"

METRICS_PREFIXES = {
    "prom433_battery_ok": [BATTERY_HELP, BATTERY_TYPE],
    "prom433_battery_V": [BATTERY_V_HELP, BATTERY_V_TYPE],
    "prom433_supercap_V": [SUPERCAP_V_HELP, SUPERCAP_V_TYPE],
    "prom433_temperature": [TEMP_HELP, TEMP_TYPE],
    "prom433_humidity": [HUMIDITY_HELP, HUMIDITY_TYPE],
    "prom433_wind_dir_deg": [WIND_DIR_HELP, WIND_DIR_TYPE],
    "prom433_wind_avg": [WIND_AVG_HELP, WIND_AVG_TYPE],
    "prom433_wind_max": [WIND_MAX_HELP, WIND_MAX_TYPE],
    "prom433_wind_avg_m": [WIND_AVG_HELP_M, WIND_AVG_TYPE],
    "prom433_wind_max_m": [WIND_MAX_HELP_M, WIND_MAX_TYPE],
    "prom433_rain": [RAIN_HELP, RAIN_TYPE],
    "prom433_uv": [UV_HELP, UV_TYPE],
    "prom433_uvi": [UVI_HELP, UVI_TYPE],
    "prom433_light_lux": [LUX_HELP, LUX_TYPE],
    "prom433_last_message": [LAST_MESSAGE_HELP, LAST_MESSAGE_TYPE],
    "prom433_freq": [FREQ_HELP, FREQ_TYPE],
    "prom433_freq1": [FREQ_HELP, FREQ_TYPE],
    "prom433_freq2": [FREQ_HELP, FREQ_TYPE],
    "prom433_rssi": [RSSI_HELP, RSSI_TYPE],
    "prom433_snr": [SNR_HELP, SNR_TYPE],
    "prom433_noise": [NOISE_HELP, NOISE_TYPE],
    "prom433_radio_clock": [RADIO_CLOCK_HELP, RADIO_CLOCK_TYPE],
    "prom433_firmware": [FIRMWARE_HELP, FIRMWARE_TYPE]
}

METRICS_CONVERT = {
    "prom433_radio_clock":
        lambda x: dateutil.utils.default_tzinfo(dateutil.parser.parse(x),
                                                dateutil.tz.tzoffset("UTC", 0))
        .timestamp(),
    # TODO: need to only do if the original metric is in mV
    "prom433_battery_V": lambda v: v / 1000.0,
    "prom433_firmware":
        lambda v: v if isinstance(v, int) or v.isnumeric() else None
}

TAG_KEYS = {"id", "channel", "model"}

IGNORE_TAGS = {
    "*": {"mic", "mod"},
    "Fineoffset-WHx080": {"subtype"},
    "LaCrosse-TX35DTHIT": {"newbattery"},
    "LaCrosse-TX29IT": {"newbattery"},
    "Fineoffset-WS90": {"flags", "data"}
}

METRIC_NAME = {
    "battery_ok": "prom433_battery_ok",
    "battery_mV": "prom433_battery_V",
    "supercap_V": "prom433_supercap_V",
    "temperature_C": "prom433_temperature",
    "humidity": "prom433_humidity",
    "wind_dir_deg": "prom433_wind_dir_deg",
    "wind_avg_km_h": "prom433_wind_avg",
    "wind_max_km_h": "prom433_wind_max",
    "wind_avg_m_s": "prom433_wind_avg_m",
    "wind_max_m_s": "prom433_wind_max_m",
    "rain_mm": "prom433_rain",
    "uv": "prom433_uv",
    "uvi": "prom433_uvi",
    "light_lux": "prom433_light_lux",
    "last_message": "prom433_last_message",
    "freq": "prom433_freq",
    "freq1": "prom433_freq1",
    "freq2": "prom433_freq2",
    "rssi": "prom433_rssi",
    "snr": "prom433_snr",
    "noise": "prom433_noise",
    "radio_clock": "prom433_radio_clock",
    "light_lux": "prom433_light_lux",
    "firmware": "prom433_firmware",
}

# {"time" : "2021-05-08 15:27:58", "model" : "Fineoffset-WHx080",
# "subtype" : 0, "id" : 202, "battery_ok" : 0, "temperature_C" : 6.900,
# "humidity" : 63, "wind_dir_deg" : 158, "wind_avg_km_h" : 4.896,
# "wind_max_km_h" : 8.568, "rain_mm" : 2.400, "mic" : "CRC"}
# {"time" : "2021-05-08 15:28:02", "model" : "Nexus-TH", "id" : 177,
# "channel" : 3, "battery_ok" : 0, "temperature_C" : 21.300, "humidity" : 39}
# {"time":"2022-11-08 14:54:14","model":"Eurochron-EFTH800",
# "id":3672,"channel":6,"battery_ok":1,"temperature_C":20.5,"humidity":60,"mic":"CRC",
# "mod":"ASK","freq":433.91904,"rssi":-0.117409,"snr":20.23702,"noise":-20.3544}


def prometheus(message, drop_after):
    payload = json.loads(message)

    tags, data, unknown = {}, {}, {}

    for key, value in payload.items():
        if key == "time":
            if ':' in payload[key]:
                time_value = dateutil.parser.parse(payload[key]).timestamp()
            else:
                time_value = datetime.fromtimestamp(float(payload[key])) \
                             .timestamp()
        elif key in TAG_KEYS:
            tags[key] = value
        elif key in METRIC_NAME:
            data[key] = value
        else:
            unknown[key] = value

    tag_value = ", ".join(["%s=\"%s\"" % (k, payload[k])
                           for k in sorted(tags)])

    METRICS["prom433_last_message"][tag_value] = time_value
    for key in data:
        metric = METRIC_NAME[key]
        if metric is None:
            continue
        if metric not in METRICS:
            METRICS[metric] = {}
        METRICS[metric][tag_value] = \
            METRICS_CONVERT.get(metric, lambda x: x)(payload[key])

    unknown = {key: value for (key, value) in unknown.items() if key not in
               (IGNORE_TAGS["*"] | IGNORE_TAGS.get(tags["model"], set()))}

    if len(unknown) > 0:
        logging.warn(f"Message has unknown tags ({unknown}): {message}")

    if drop_after > 0:
        tags_to_drop = set()

        for tag_value in METRICS["prom433_last_message"].keys():
            limit = time_value - drop_after
            if METRICS["prom433_last_message"][tag_value] < limit:
                tags_to_drop.add(tag_value)

        if len(tags_to_drop) == 0:
            return

        logging.info(f"Dropping {len(tags_to_drop)} as not seen"
                     + " for {drop_after} seconds")
        for tag_value in tags_to_drop:
            logging.info(f"Dropping {tag_value}")
            for metric in METRICS.keys():
                if tag_value in METRICS[metric]:
                    del METRICS[metric][tag_value]


def get_metrics():
    lines = []
    for metric_name in sorted(METRICS.keys()):
        lines.append(HELP_FORMAT
                     % (metric_name, METRICS_PREFIXES[metric_name][0]))
        lines.append(TYPE_FORMAT
                     % (metric_name, METRICS_PREFIXES[metric_name][1]))
        for (tags, values) in METRICS[metric_name].items():
            if values is not None:
                lines.append(METRIC_FORMAT % (metric_name, tags, values))

    return "\n".join(lines)



# glowprom
# Copyright (C) 2020 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
import http.server
import traceback

from sentry_sdk import capture_exception  # type:ignore

from .prometheus import get_metrics


def handle_error(func):
    def r(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except Exception as e:
            traceback.print_exception(e)
            capture_exception(e)

            self.send_response(500)
            self.send_header("Content-type", "text/plain")
            self.end_headers()

            self.wfile.write(f"Exception Occurred.\n".encode("utf8"))
    return r


class Handler(http.server.BaseHTTPRequestHandler):
    def __init__(self, quiet=False):
        self.quiet = quiet

    def __call__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def do_GET(self):
        if self.path == "/":
            self.send_index()
        elif self.path == "/metrics":
            self.send_metrics()
        else:
            self.send_error(404)

    def send_index(self):
        self.send_response(200)
        self.end_headers()
        self.wfile.write("""
<html>
<head><title>RTL433 Prometheus</title></head>
<body>
<h1>RTL433 Prometheus</h1>
<p><a href="/metrics">Metrics</a></p>
</body>
</html>""".encode("utf8"))

    def send_metrics(self):
        self.send_response(200)
        self.end_headers()
        self.wfile.write(get_metrics().encode("utf8"))

    def log_request(self, code='-', size='-'):
        if self.quiet:
            return
        else:
            self.log_message('"%s" %s %s',
                             self.requestline, str(code), str(size))


def serve(args):  # pragma: no cover
    handler = Handler(args.quiet)
    server = http.server.HTTPServer(args.bind, handler)
    server.serve_forever()



# prom433
# Copyright (C) 2021 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import os
import unittest

from prom433 import prometheus, get_metrics
from prom433.prometheus import METRICS

MESSAGE_TEXT = open("tests/output_sample.txt", "rb").read().decode("utf8")
DROP_TEXT = open("tests/dropmetric_sample.txt", "rb").read().decode("utf8")
TIMESTAMP_TEXT = open("tests/timestamp_sample.txt", "rb").read().decode("utf8")


def mock_popen(args, stdout):
    return MockPopenReturn(io.StringIO(MESSAGE_TEXT))


class MockPopenReturn:
    def __init__(self, buffer):
        self.stdout = buffer


class TestPrometheus(unittest.TestCase):
    def test_prometheus(self):
        for line in MESSAGE_TEXT.split("\n"):
            prometheus(line, 0)

        prom = get_metrics()

        self.assertIn(
            """prom433_temperature{channel="1", id="147\"""" +
            """, model="Nexus-TH"} 23.100000""", prom)
        self.assertIn(
            """prom433_temperature{id="250", """ +
            """model="Fineoffset-WHx080"} 16.000000""",
            prom)
        self.assertIn(
            """prom433_temperature{channel="2", id="1940", """ +
            """model="Eurochron-EFTH800"} 22.300000""", prom)
        self.assertIn(
            """prom433_noise{channel="6", id="3672", """ +
            """model="Eurochron-EFTH800"} -20.3544""", prom)
        self.assertIn(
            """prom433_radio_clock{channel="2", id="1940", """ +
            """model="Eurochron-EFTH800"} 1670453240.000000""", prom)
        self.assertIn(
            """prom433_battery_V{id="18326", model="Fineoffset-WS90"} """ +
            """2.660000""", prom)

    def test_drop_metric_after(self):
        for line in DROP_TEXT.split("\n"):
            prometheus(line, 3600)

        prom = get_metrics()

        # The Nexus wasn't dropped because it was updated 30 minutes ago,
        self.assertIn("""Nexus-TH""", prom)
        # The Eurochron message was the one that triggered the delete...
        self.assertIn("""Eurochron-EFTH800""", prom)
        # but the Fineoffset hasn't been seen for an hour, so was dropped.
        self.assertNotIn("""Fineoffset-WHx080""", prom)

    def test_drop_metric_disabled(self):
        for line in DROP_TEXT.split("\n"):
            prometheus(line, 0)

        prom = get_metrics()

        # The Nexus wasn't dropped because it was updated 30 minutes ago,
        self.assertIn("""Nexus-TH""", prom)
        # The Eurochron message was the one that triggered the delete...
        self.assertIn("""Eurochron-EFTH800""", prom)
        # The Fineoffset hasn't been seen for an hour,
        # but dropping is disabled.
        self.assertIn("""Fineoffset-WHx080""", prom)

    def test_timestamp(self):
        for line in TIMESTAMP_TEXT.split("\n"):
            prometheus(line, 0)

        prom = get_metrics()

        # Test for time in format time:utc:tz
        self.assertIn(
            """prom433_last_message{id="1", """ +
            """model="LaCrosse-TX"} 1677374905.000000""", prom)
        # Test for time in format time:tz
        self.assertIn(
            """prom433_last_message{id="2", """ +
            """model="LaCrosse-TX"} 1677374905.000000""", prom)
        # Test for time in format time:iso:tz
        self.assertIn(
            """prom433_last_message{id="3", """ +
            """model="LaCrosse-TX"} 1677374905.000000""", prom)
        # Test for time in format time:unix
        self.assertIn(
            """prom433_last_message{id="4", """ +
            """model="LaCrosse-TX"} 1677374905.000000""", prom)
        # Test for time in format time:utc:tz:usec
        self.assertIn(
            """prom433_last_message{id="101", """ +
            """model="LaCrosse-TX31UIT"} 1677374905.538138""", prom)
        # Test for time in format time:tz:usec
        self.assertIn(
            """prom433_last_message{id="102", """ +
            """model="LaCrosse-TX31UIT"} 1677374905.538138""", prom)
        # Test for time in format time:iso:tz:usec
        self.assertIn(
            """prom433_last_message{id="103", """ +
            """model="LaCrosse-TX31UIT"} 1677374905.538138""", prom)
        # Test for time in format time:unix:usec
        self.assertIn(
            """prom433_last_message{id="104", """ +
            """model="LaCrosse-TX31UIT"} 1677374905.538138""", prom)



# prom433
# Copyright (C) 2021 Andrew Wilkinson
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import os

parser = argparse.ArgumentParser(
    description='Listens to messages from rtl_433 and exposes them '
    + 'as prometheus metrics')
parser.add_argument('-q', '--quiet', action="store_true",
                    help="don't log HTTP requests")
parser.add_argument('--bind', type=str, nargs='?', default="0.0.0.0:9100",
                    help='the ip address and port to bind to. Default: *:9100')
parser.add_argument('--mqtt', type=str, nargs='?', default="mqtt",
                    help="the mqtt host to connect to. Default: mqtt:1883")
parser.add_argument('--drop-after', type=int, nargs='?', default=3600,
                    help="drop metrics this many seconds after"
                         + " the device was last seen."
                         + " 0 disables dropping metrics")


def get_arguments(args):
    args = parser.parse_args(args)
    if "MQTT_HOST" in os.environ:
        args.mqtt = os.environ["MQTT_HOST"]

    if ":" not in args.bind:
        args.bind = (args.bind, 9100)
    else:
        args.bind = (args.bind.split(":")[0], int(args.bind.split(":")[1]))

    return args
